{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "# For tips on running notebooks in Google Colab, see\n# https://pytorch.org/tutorials/beginner/colab\n%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Saving and Loading Models\n**Author:** [Matthew Inkawhich](https://github.com/MatthewInkawhich)\n\nThis document provides solutions to a variety of use cases regarding the\nsaving and loading of PyTorch models. Feel free to read the whole\ndocument, or just skip to the code you need for a desired use case.\n\nWhen it comes to saving and loading models, there are three core\nfunctions to be familiar with:\n\n1) [torch.save](https://pytorch.org/docs/stable/torch.html?highlight=save#torch.save)_:\n   Saves a serialized object to disk. This function uses Python\u2019s\n   [pickle](https://docs.python.org/3/library/pickle.html)_ utility\n   for serialization. Models, tensors, and dictionaries of all kinds of\n   objects can be saved using this function.\n\n2) [torch.load](https://pytorch.org/docs/stable/torch.html?highlight=torch%20load#torch.load)_:\n   Uses [pickle](https://docs.python.org/3/library/pickle.html)_\\ \u2019s\n   unpickling facilities to deserialize pickled object files to memory.\n   This function also facilitates the device to load the data into (see\n   [Saving & Loading Model Across\n   Devices](#saving-loading-model-across-devices)_).\n\n3) [torch.nn.Module.load_state_dict](https://pytorch.org/docs/stable/generated/torch.nn.Module.html?highlight=load_state_dict#torch.nn.Module.load_state_dict)_:\n   Loads a model\u2019s parameter dictionary using a deserialized\n   *state_dict*. For more information on *state_dict*, see [What is a\n   state_dict?](#what-is-a-state-dict)_.\n\n\n\n**Contents:**\n\n-  [What is a state_dict?](#what-is-a-state-dict)_\n-  [Saving & Loading Model for\n   Inference](#saving-loading-model-for-inference)_\n-  [Saving & Loading a General\n   Checkpoint](#saving-loading-a-general-checkpoint-for-inference-and-or-resuming-training)_\n-  [Saving Multiple Models in One\n   File](#saving-multiple-models-in-one-file)_\n-  [Warmstarting Model Using Parameters from a Different\n   Model](#warmstarting-model-using-parameters-from-a-different-model)_\n-  [Saving & Loading Model Across\n   Devices](#saving-loading-model-across-devices)_\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## What is a ``state_dict``?\n\nIn PyTorch, the learnable parameters (i.e.\u00a0weights and biases) of an\n``torch.nn.Module`` model are contained in the model\u2019s *parameters*\n(accessed with ``model.parameters()``). A *state_dict* is simply a\nPython dictionary object that maps each layer to its parameter tensor.\nNote that only layers with learnable parameters (convolutional layers,\nlinear layers, etc.) and registered buffers (batchnorm's running_mean)\nhave entries in the model\u2019s *state_dict*. Optimizer\nobjects (``torch.optim``) also have a *state_dict*, which contains\ninformation about the optimizer's state, as well as the hyperparameters\nused.\n\nBecause *state_dict* objects are Python dictionaries, they can be easily\nsaved, updated, altered, and restored, adding a great deal of modularity\nto PyTorch models and optimizers.\n\n### Example:\n\nLet\u2019s take a look at the *state_dict* from the simple model used in the\n[Training a\nclassifier](https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html#sphx-glr-beginner-blitz-cifar10-tutorial-py)_\ntutorial.\n\n.. code:: python\n\n   # Define model\n   class TheModelClass(nn.Module):\n       def __init__(self):\n           super(TheModelClass, self).__init__()\n           self.conv1 = nn.Conv2d(3, 6, 5)\n           self.pool = nn.MaxPool2d(2, 2)\n           self.conv2 = nn.Conv2d(6, 16, 5)\n           self.fc1 = nn.Linear(16 * 5 * 5, 120)\n           self.fc2 = nn.Linear(120, 84)\n           self.fc3 = nn.Linear(84, 10)\n\n       def forward(self, x):\n           x = self.pool(F.relu(self.conv1(x)))\n           x = self.pool(F.relu(self.conv2(x)))\n           x = x.view(-1, 16 * 5 * 5)\n           x = F.relu(self.fc1(x))\n           x = F.relu(self.fc2(x))\n           x = self.fc3(x)\n           return x\n\n   # Initialize model\n   model = TheModelClass()\n\n   # Initialize optimizer\n   optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.9)\n\n   # Print model's state_dict\n   print(\"Model's state_dict:\")\n   for param_tensor in model.state_dict():\n       print(param_tensor, \"\\t\", model.state_dict()[param_tensor].size())\n\n   # Print optimizer's state_dict\n   print(\"Optimizer's state_dict:\")\n   for var_name in optimizer.state_dict():\n       print(var_name, \"\\t\", optimizer.state_dict()[var_name])\n\n**Output:**\n\n::\n\n   Model's state_dict:\n   conv1.weight     torch.Size([6, 3, 5, 5])\n   conv1.bias   torch.Size([6])\n   conv2.weight     torch.Size([16, 6, 5, 5])\n   conv2.bias   torch.Size([16])\n   fc1.weight   torch.Size([120, 400])\n   fc1.bias     torch.Size([120])\n   fc2.weight   torch.Size([84, 120])\n   fc2.bias     torch.Size([84])\n   fc3.weight   torch.Size([10, 84])\n   fc3.bias     torch.Size([10])\n\n   Optimizer's state_dict:\n   state    {}\n   param_groups     [{'lr': 0.001, 'momentum': 0.9, 'dampening': 0, 'weight_decay': 0, 'nesterov': False, 'params': [4675713712, 4675713784, 4675714000, 4675714072, 4675714216, 4675714288, 4675714432, 4675714504, 4675714648, 4675714720]}]\n\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Saving & Loading Model for Inference\n\n### Save/Load ``state_dict`` (Recommended)\n\n**Save:**\n\n.. code:: python\n\n   torch.save(model.state_dict(), PATH)\n\n**Load:**\n\n.. code:: python\n\n   model = TheModelClass(*args, **kwargs)\n   model.load_state_dict(torch.load(PATH))\n   model.eval()\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>The 1.6 release of PyTorch switched ``torch.save`` to use a new\n    zip file-based format. ``torch.load`` still retains the ability to\n    load files in the old format. If for any reason you want ``torch.save``\n    to use the old format, pass the ``kwarg``parameter ``_use_new_zipfile_serialization=False``.</p></div>\n\nWhen saving a model for inference, it is only necessary to save the\ntrained model\u2019s learned parameters. Saving the model\u2019s *state_dict* with\nthe ``torch.save()`` function will give you the most flexibility for\nrestoring the model later, which is why it is the recommended method for\nsaving models.\n\nA common PyTorch convention is to save models using either a ``.pt`` or\n``.pth`` file extension.\n\nRemember that you must call ``model.eval()`` to set dropout and batch\nnormalization layers to evaluation mode before running inference.\nFailing to do this will yield inconsistent inference results.\n\n.. Note ::\n\n   Notice that the ``load_state_dict()`` function takes a dictionary\n   object, NOT a path to a saved object. This means that you must\n   deserialize the saved *state_dict* before you pass it to the\n   ``load_state_dict()`` function. For example, you CANNOT load using\n   ``model.load_state_dict(PATH)``.\n\n.. Note ::\n\n   If you only plan to keep the best performing model (according to the \n   acquired validation loss), don't forget that ``best_model_state = model.state_dict()``\n   returns a reference to the state and not its copy! You must serialize \n   ``best_model_state`` or use ``best_model_state = deepcopy(model.state_dict())`` otherwise\n   your best ``best_model_state`` will keep getting updated by the subsequent training \n   iterations. As a result, the final model state will be the state of the overfitted model. \n\n### Save/Load Entire Model\n\n**Save:**\n\n.. code:: python\n\n   torch.save(model, PATH)\n\n**Load:**\n\n.. code:: python\n\n   # Model class must be defined somewhere\n   model = torch.load(PATH)\n   model.eval()\n\nThis save/load process uses the most intuitive syntax and involves the\nleast amount of code. Saving a model in this way will save the entire\nmodule using Python\u2019s\n[pickle](https://docs.python.org/3/library/pickle.html)_ module. The\ndisadvantage of this approach is that the serialized data is bound to\nthe specific classes and the exact directory structure used when the\nmodel is saved. The reason for this is because pickle does not save the\nmodel class itself. Rather, it saves a path to the file containing the\nclass, which is used during load time. Because of this, your code can\nbreak in various ways when used in other projects or after refactors.\n\nA common PyTorch convention is to save models using either a ``.pt`` or\n``.pth`` file extension.\n\nRemember that you must call ``model.eval()`` to set dropout and batch\nnormalization layers to evaluation mode before running inference.\nFailing to do this will yield inconsistent inference results.\n\n### Export/Load Model in TorchScript Format\n\nOne common way to do inference with a trained model is to use\n[TorchScript](https://pytorch.org/docs/stable/jit.html)_, an intermediate\nrepresentation of a PyTorch model that can be run in Python as well as in a\nhigh performance environment like C++. TorchScript is actually the recommended model format\nfor scaled inference and deployment.\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>Using the TorchScript format, you will be able to load the exported model and\n   run inference without defining the model class.</p></div>\n\n**Export:**\n\n.. code:: python\n\n   model_scripted = torch.jit.script(model) # Export to TorchScript\n   model_scripted.save('model_scripted.pt') # Save\n\n**Load:**\n\n.. code:: python\n\n   model = torch.jit.load('model_scripted.pt')\n   model.eval()\n\nRemember that you must call ``model.eval()`` to set dropout and batch\nnormalization layers to evaluation mode before running inference.\nFailing to do this will yield inconsistent inference results.\n\nFor more information on TorchScript, feel free to visit the dedicated\n[tutorials](https://pytorch.org/tutorials/beginner/Intro_to_TorchScript_tutorial.html)_.\nYou will get familiar with the tracing conversion and learn how to\nrun a TorchScript module in a [C++ environment](https://pytorch.org/tutorials/advanced/cpp_export.html)_.\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Saving & Loading a General Checkpoint for Inference and/or Resuming Training\n\n### Save:\n\n.. code:: python\n\n   torch.save({\n               'epoch': epoch,\n               'model_state_dict': model.state_dict(),\n               'optimizer_state_dict': optimizer.state_dict(),\n               'loss': loss,\n               ...\n               }, PATH)\n\n### Load:\n\n.. code:: python\n\n   model = TheModelClass(*args, **kwargs)\n   optimizer = TheOptimizerClass(*args, **kwargs)\n\n   checkpoint = torch.load(PATH)\n   model.load_state_dict(checkpoint['model_state_dict'])\n   optimizer.load_state_dict(checkpoint['optimizer_state_dict'])\n   epoch = checkpoint['epoch']\n   loss = checkpoint['loss']\n\n   model.eval()\n   # - or -\n   model.train()\n\nWhen saving a general checkpoint, to be used for either inference or\nresuming training, you must save more than just the model\u2019s\n*state_dict*. It is important to also save the optimizer's *state_dict*,\nas this contains buffers and parameters that are updated as the model\ntrains. Other items that you may want to save are the epoch you left off\non, the latest recorded training loss, external ``torch.nn.Embedding``\nlayers, etc. As a result, such a checkpoint is often 2~3 times larger \nthan the model alone.\n\nTo save multiple components, organize them in a dictionary and use\n``torch.save()`` to serialize the dictionary. A common PyTorch\nconvention is to save these checkpoints using the ``.tar`` file\nextension.\n\nTo load the items, first initialize the model and optimizer, then load\nthe dictionary locally using ``torch.load()``. From here, you can easily\naccess the saved items by simply querying the dictionary as you would\nexpect.\n\nRemember that you must call ``model.eval()`` to set dropout and batch\nnormalization layers to evaluation mode before running inference.\nFailing to do this will yield inconsistent inference results. If you\nwish to resuming training, call ``model.train()`` to ensure these layers\nare in training mode.\n\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Saving Multiple Models in One File\n\n### Save:\n\n.. code:: python\n\n   torch.save({\n               'modelA_state_dict': modelA.state_dict(),\n               'modelB_state_dict': modelB.state_dict(),\n               'optimizerA_state_dict': optimizerA.state_dict(),\n               'optimizerB_state_dict': optimizerB.state_dict(),\n               ...\n               }, PATH)\n\n### Load:\n\n.. code:: python\n\n   modelA = TheModelAClass(*args, **kwargs)\n   modelB = TheModelBClass(*args, **kwargs)\n   optimizerA = TheOptimizerAClass(*args, **kwargs)\n   optimizerB = TheOptimizerBClass(*args, **kwargs)\n\n   checkpoint = torch.load(PATH)\n   modelA.load_state_dict(checkpoint['modelA_state_dict'])\n   modelB.load_state_dict(checkpoint['modelB_state_dict'])\n   optimizerA.load_state_dict(checkpoint['optimizerA_state_dict'])\n   optimizerB.load_state_dict(checkpoint['optimizerB_state_dict'])\n\n   modelA.eval()\n   modelB.eval()\n   # - or -\n   modelA.train()\n   modelB.train()\n\nWhen saving a model comprised of multiple ``torch.nn.Modules``, such as\na GAN, a sequence-to-sequence model, or an ensemble of models, you\nfollow the same approach as when you are saving a general checkpoint. In\nother words, save a dictionary of each model\u2019s *state_dict* and\ncorresponding optimizer. As mentioned before, you can save any other\nitems that may aid you in resuming training by simply appending them to\nthe dictionary.\n\nA common PyTorch convention is to save these checkpoints using the\n``.tar`` file extension.\n\nTo load the models, first initialize the models and optimizers, then\nload the dictionary locally using ``torch.load()``. From here, you can\neasily access the saved items by simply querying the dictionary as you\nwould expect.\n\nRemember that you must call ``model.eval()`` to set dropout and batch\nnormalization layers to evaluation mode before running inference.\nFailing to do this will yield inconsistent inference results. If you\nwish to resuming training, call ``model.train()`` to set these layers to\ntraining mode.\n\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Warmstarting Model Using Parameters from a Different Model\n\n### Save:\n\n.. code:: python\n\n   torch.save(modelA.state_dict(), PATH)\n\n### Load:\n\n.. code:: python\n\n   modelB = TheModelBClass(*args, **kwargs)\n   modelB.load_state_dict(torch.load(PATH), strict=False)\n\nPartially loading a model or loading a partial model are common\nscenarios when transfer learning or training a new complex model.\nLeveraging trained parameters, even if only a few are usable, will help\nto warmstart the training process and hopefully help your model converge\nmuch faster than training from scratch.\n\nWhether you are loading from a partial *state_dict*, which is missing\nsome keys, or loading a *state_dict* with more keys than the model that\nyou are loading into, you can set the ``strict`` argument to **False**\nin the ``load_state_dict()`` function to ignore non-matching keys.\n\nIf you want to load parameters from one layer to another, but some keys\ndo not match, simply change the name of the parameter keys in the\n*state_dict* that you are loading to match the keys in the model that\nyou are loading into.\n\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Saving & Loading Model Across Devices\n\n### Save on GPU, Load on CPU\n\n**Save:**\n\n.. code:: python\n\n   torch.save(model.state_dict(), PATH)\n\n**Load:**\n\n.. code:: python\n\n   device = torch.device('cpu')\n   model = TheModelClass(*args, **kwargs)\n   model.load_state_dict(torch.load(PATH, map_location=device))\n\nWhen loading a model on a CPU that was trained with a GPU, pass\n``torch.device('cpu')`` to the ``map_location`` argument in the\n``torch.load()`` function. In this case, the storages underlying the\ntensors are dynamically remapped to the CPU device using the\n``map_location`` argument.\n\n### Save on GPU, Load on GPU\n\n**Save:**\n\n.. code:: python\n\n   torch.save(model.state_dict(), PATH)\n\n**Load:**\n\n.. code:: python\n\n   device = torch.device(\"cuda\")\n   model = TheModelClass(*args, **kwargs)\n   model.load_state_dict(torch.load(PATH))\n   model.to(device)\n   # Make sure to call input = input.to(device) on any input tensors that you feed to the model\n\nWhen loading a model on a GPU that was trained and saved on GPU, simply\nconvert the initialized ``model`` to a CUDA optimized model using\n``model.to(torch.device('cuda'))``. Also, be sure to use the\n``.to(torch.device('cuda'))`` function on all model inputs to prepare\nthe data for the model. Note that calling ``my_tensor.to(device)``\nreturns a new copy of ``my_tensor`` on GPU. It does NOT overwrite\n``my_tensor``. Therefore, remember to manually overwrite tensors:\n``my_tensor = my_tensor.to(torch.device('cuda'))``.\n\n### Save on CPU, Load on GPU\n\n**Save:**\n\n.. code:: python\n\n   torch.save(model.state_dict(), PATH)\n\n**Load:**\n\n.. code:: python\n\n   device = torch.device(\"cuda\")\n   model = TheModelClass(*args, **kwargs)\n   model.load_state_dict(torch.load(PATH, map_location=\"cuda:0\"))  # Choose whatever GPU device number you want\n   model.to(device)\n   # Make sure to call input = input.to(device) on any input tensors that you feed to the model\n\nWhen loading a model on a GPU that was trained and saved on CPU, set the\n``map_location`` argument in the ``torch.load()`` function to\n``cuda:device_id``. This loads the model to a given GPU device. Next, be\nsure to call ``model.to(torch.device('cuda'))`` to convert the model\u2019s\nparameter tensors to CUDA tensors. Finally, be sure to use the\n``.to(torch.device('cuda'))`` function on all model inputs to prepare\nthe data for the CUDA optimized model. Note that calling\n``my_tensor.to(device)`` returns a new copy of ``my_tensor`` on GPU. It\ndoes NOT overwrite ``my_tensor``. Therefore, remember to manually\noverwrite tensors: ``my_tensor = my_tensor.to(torch.device('cuda'))``.\n\n### Saving ``torch.nn.DataParallel`` Models\n\n**Save:**\n\n.. code:: python\n\n   torch.save(model.module.state_dict(), PATH)\n\n**Load:**\n\n.. code:: python\n\n   # Load to whatever device you want\n\n``torch.nn.DataParallel`` is a model wrapper that enables parallel GPU\nutilization. To save a ``DataParallel`` model generically, save the\n``model.module.state_dict()``. This way, you have the flexibility to\nload the model any way you want to any device you want.\n\n\n"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.10.12"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}